import logging

from ..base import BaseSource

logger = logging.getLogger(__name__)

PREFIXES = ["", "/heads", "/tags"]


class BasicGitHubSource(BaseSource):
    SOURCE_KEY = "basicgithub"

    def __init__(
        self,
        owner: str,
        repo: str,
        ref=None,
        user_agent="curl/7.83.1",
        file_type="tar.gz",
        rehash_if_same_url=False,
        **kwargs,
    ):
        self.owner = owner
        self.repo = repo
        self.ref = ref
        self.headers = {"User-Agent": user_agent}
        self.user_agent = user_agent
        self.file_type = file_type
        self.rehash_if_same_url = rehash_if_same_url

        assert file_type in [
            "zip",
            "tar.gz",
        ], "file type must be either zip or tar.gz"

    def _get_default_branch(self):
        from json import load
        from urllib import request

        url = f"https://api.github.com/repos/{self.owner}/{self.repo}"
        res = request.urlopen(request.Request(url, headers=self.headers))
        jres = load(res)
        logger.debug(url, jres)
        branch = jres["default_branch"]
        return f"heads/{branch}"

    def reduce(self, **kwargs):
        from json import load
        from urllib import request

        logger.debug(
            dict(
                owner=self.owner,
                repo=self.repo,
                ref=self.ref,
                file_type=self.file_type,
            )
        )  # noqa: E501

        ret = kwargs

        if self.ref is None:
            if ret.get("default_branch") is None:
                ret["default_branch"] = self._get_default_branch()
            self.ref = ret["default_branch"]

        assert self.ref is not None, "ref must not be None"

        self.commit_id = None
        for prefix in PREFIXES:
            try:
                url = f"https://api.github.com/repos/{self.owner}/{self.repo}/git/matching-refs{prefix}/{self.ref}"  # noqa: E501
                logger.debug(url)
                res = request.urlopen(
                    request.Request(url, headers=self.headers)
                )
                jres = load(res)
                if len(jres) > 0:
                    obj = jres[0]["object"]
                    if obj["type"] != "commit":
                        res = request.urlopen(
                            request.Request(obj["url"], headers=self.headers)
                        )
                        jres = load(res)
                        self.commit_id = jres["object"]["sha"]
                        break
                    else:
                        self.commit_id = obj["sha"]
                        break
            except request.HTTPError as e:
                if e.code == 404:
                    continue
                raise e
        assert (
            self.commit_id is not None
        ), f"ref {self.ref} is not valid for {self.owner}/{self.repo}"
        ret["github_commit"] = self.commit_id
        logger.info(
            f"{self.owner}/{self.repo} latest github commit for ref {self.ref} is {self.commit_id}"  # noqa: E501
        )

        res = request.urlopen(
            request.Request(
                f"https://github.com/{self.owner}/{self.repo}/archive/{self.commit_id}.{self.file_type}",  # noqa: E501
                headers=self.headers,
            )
        )
        resolved_url = res.url
        logger.debug(
            dict(
                url=resolved_url,
                rehash_if_same_url=self.rehash_if_same_url,
                user_agent=self.user_agent,
            )
        )

        if resolved_url != ret.get("final_url") or self.rehash_if_same_url:
            logger.info(f"Downloading and hashing: {resolved_url}")
            import hashlib

            hasher = hashlib.sha256()
            while True:
                buf = res.read(16 * 1024)
                if not buf:
                    break
                hasher.update(buf)
            ret["sha256"] = hasher.hexdigest()
        ret["final_url"] = res.url
        ret["file_type"] = self.file_type
        return ret

    @classmethod
    def argparse(cls, parser):
        parser.description = "Basic fetcher for GitHub"
        parser.add_argument("owner", type=str)
        parser.add_argument("repo", type=str)
        parser.add_argument("-b,--ref", type=str)
        parser.add_argument(
            "-t,--file-type", choices=["tar.gz", "zip"], default="tar.gz"
        )
        parser.add_argument("-r,--rehash-if-same-url", action="store_true")
        return parser



class BaseSource:
    SOURCE_KEY = "_base"

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def reduce(self, **kwargs):
        raise Exception("Unimplemented")

    @classmethod
    def argparse(cls, parser):
        raise Exception("Unimplemented")
        return parser



import logging

from ..base import BaseSource

logger = logging.getLogger(__name__)


class BasicHTTPSource(BaseSource):
    SOURCE_KEY = "basichttp"

    def __init__(
        self,
        url: str,
        rehash_if_same_url=False,
        user_agent="curl/7.83.1",
        **kwargs,
    ):
        self.url = url
        self.user_agent = user_agent
        self.rehash_if_same_url = rehash_if_same_url

    def reduce(self, **kwargs):
        from urllib import request

        ret = kwargs
        res = request.urlopen(
            request.Request(self.url, headers={"User-Agent": self.user_agent})
        )
        resolved_url = res.url
        logger.debug(
            dict(
                url=self.url,
                rehash_if_same_url=self.rehash_if_same_url,
                user_agent=self.user_agent,
            )
        )

        if resolved_url != ret.get("final_url") or self.rehash_if_same_url:
            logger.info(f"Downloading and hashing: {resolved_url}")
            import hashlib

            hasher = hashlib.sha256()
            while True:
                buf = res.read(16 * 1024)
                if not buf:
                    break
                hasher.update(buf)
            ret["sha256"] = hasher.hexdigest()
        ret["final_url"] = res.url
        return ret

    @classmethod
    def argparse(cls, parser):
        parser.description = "Basic fetcher for HTTP"
        parser.add_argument("url", type=str)
        parser.add_argument("-r,--rehash-if-same-url", action="store_true")
        return parser



import logging

logger = logging.getLogger(__name__)
logging.basicConfig()

"""CLI interface for bumpkin project.

Be creative! do whatever you want!

- Install click or typer and create a CLI app
- Use builtin argparse
- Start a web application
- Import things from your .base module
"""


def demo_subcommand(subparser):
    subparser.description = (
        "Execute specific bump modules with specified arguments"
    )
    from .sources import get_subcommands

    get_subcommands(subparser.add_subparsers())


def eval_subcommand(subparser):
    from pathlib import Path

    subparser.description = "Evaluate a bumpkin formatted JSON to an output"
    subparser.add_argument("-v,--verbose", dest="verbose", action="store_true")
    subparser.add_argument(
        "-i,--input", dest="input_file", type=Path, required=True
    )
    subparser.add_argument("-o,--output", dest="output_file", type=Path)
    subparser.add_argument(
        "-p,--pretty",
        dest="indent",
        help="Enable JSON identations instead of minified",
        action="store_true",
    )
    subparser.add_argument(
        dest="keys",
        nargs="*",
        type=str,
        help="Bump only these keys. If ommited, bump all.",
    )

    def handle(input_file, keys, indent, output_file=None, **kwargs):
        from json import dumps, loads

        from .sources import eval_nodes

        assert input_file.exists(), f"'{input_file.resolve()}' does not exist"
        input_file_data = loads(input_file.read_text())

        if output_file is None:
            output_file = Path(str(input_file) + ".lock")
        output_file_data = None
        if output_file.exists():
            output_file_data = loads(output_file.read_text())
        processed = eval_nodes(input_file_data, output_file_data, keys)
        output_file.write_text(dumps(processed, indent=4 if indent else None))

    subparser.set_defaults(fn=handle)
    # todo implement


def list_subcommand(subparser):
    from pathlib import Path

    subparser.description = "List bumpable nodes in JSON"
    subparser.add_argument("-v,--verbose", dest="verbose", action="store_true")
    subparser.add_argument(
        "-i,--input", dest="input_file", type=Path, required=True
    )
    subparser.add_argument("-o,--output", dest="output_file", type=Path)
    subparser.add_argument(
        "-s,--show-state", dest="show_state", action="store_true"
    )

    def handle(input_file, output_file=None, show_state=False, **kwargs):
        from json import loads

        from .sources import list_nodes

        assert input_file.exists(), f"'{input_file.resolve()}' does not exist"
        if output_file is None:
            output_file = Path(str(input_file) + ".lock")
        input_file_data = loads(input_file.read_text())
        output_file_data = None
        if output_file.exists():
            output_file_data = loads(output_file.read_text())
        nodes = list_nodes(input_file_data, output_file_data)
        node_keys = list(nodes.keys())
        node_keys.sort()
        for node in node_keys:
            if show_state:
                print(node, nodes[node])
            else:
                print(node)

    subparser.set_defaults(fn=handle)
    # todo implement


def main():  # pragma: no cover
    """
    The main function executes on commands:
    `python -m bumpkin` and `$ bumpkin `.

    This is your program's entry point.

    You can change this function to do whatever you want.
    Examples:
        * Run a test suite
        * Run a server
        * Do some other stuff
        * Run a command line application (Click, Typer, ArgParse)
        * List all available tasks
        * Run an application (Flask, FastAPI, Django, etc.)
    """
    from argparse import ArgumentParser
    from sys import argv

    parser = ArgumentParser()
    subparsers = parser.add_subparsers()
    demo_subcommand(subparsers.add_parser("demo"))
    eval_subcommand(subparsers.add_parser("eval"))
    list_subcommand(subparsers.add_parser("list"))
    args = vars(parser.parse_args())

    if args.get("verbose"):
        logger.root.setLevel(logging.DEBUG)
    else:
        logger.root.setLevel(logging.INFO)

    logger.debug(f"args: {args}")

    if args.get("fn"):
        args["fn"](**args.copy())
    else:
        parser.parse_args([*argv[1:], "--help"])



import logging
from typing import Dict

from .base import BaseSource
from .basicgithub import BasicGitHubSource
from .basichttp import BasicHTTPSource
from .basichttpjsonvendor import BasicHTTPJSONVendorSource

logger = logging.getLogger(__name__)

sources: Dict[str, BaseSource] = {}

default_sources = [
    BasicHTTPSource,
    BasicGitHubSource,
    BasicHTTPJSONVendorSource,
]


def setup_source(source):
    assert issubclass(
        source, BaseSource
    ), f"class {source} is not based on BaseSource"
    assert (
        sources.get(source.SOURCE_KEY) is None
    ), f"class {source} has the samekey ({source.SOURCE_KEY}) as {sources.get(source.SOURCE_KEY)}"  # noqa: E501
    sources[source.SOURCE_KEY] = source


for default_source in default_sources:
    setup_source(default_source)


def eval_node(declaration, previous_data=dict()):
    from time import time
    from urllib import request

    assert (
        type(declaration) == dict
    ), "declaration type must be a object/dictionary"
    assert (
        type(previous_data) == dict
    ), "previous data type must be a object/dictionary"
    assert (
        type(declaration["_type"]) == str
    ), "declaration error: type of _type must be string"
    source_type = declaration["_type"]
    assert (
        sources.get(source_type) is not None
    ), f"source type {source_type} is not defined or not available"
    declaration.pop("_type")
    source = sources[source_type]
    try:
        ret = source(**declaration).reduce(**previous_data)
        ret["_bpk_last_update"] = int(time())
        return ret
    except request.HTTPError as e:
        logger.info(
            f"Unhandled HTTP error while evaluating node {declaration}, saving old state..."  # noqa: E501
        )
        logger.info(e)
    except Exception as e:
        logger.info(
            f"Unhandled generic exception while evaluating node {declaration}, saving old state"  # noqa: E501
        )
        logger.info(e)
    except KeyboardInterrupt as e:
        logger.info("Saving work and exiting...")
        raise e
    except AssertionError as e:
        logger.info(
            f"Failed assertion while evaluating node {declaration}, saving old state"  # noqa: E501
        )
        logger.info(e)

    return previous_data


def get_subcommands(subparser):
    def make_source_payload_function(source):
        def payload_fn(**kwargs):
            obj = source(**kwargs)
            print(obj.reduce())

        return payload_fn

    for source_name, source in sources.items():
        parser = subparser.add_parser(source_name)
        parser.add_argument(
            "-v,--verbose", dest="verbose", action="store_true"
        )
        parser.set_defaults(fn=make_source_payload_function(source))
        source.argparse(parser)


def list_nodes(declaration=None, previous_data=None, _key=[]):
    if type(declaration) is not dict:
        return {}
    if declaration.get("_type") is None:
        ret = {}
        for k in declaration.keys():
            key = [*_key, k]
            ret = {
                **ret,
                **list_nodes(
                    declaration.get(k),
                    previous_data.get(k)
                    if type(previous_data) is dict
                    else None,
                    key,
                ),
            }
        return ret
    else:
        ret = {}
        ret[".".join(_key)] = 1 if previous_data is not None else 0
        try:
            ret[".".join(_key)] = previous_data["_bpk_last_update"]
        except KeyError:
            pass
        except TypeError:
            pass
        return ret


def eval_nodes_recursively(declaration=None, previous_data=None):
    if type(declaration) == dict:
        if declaration.get("_type"):
            return eval_node(
                declaration,
                previous_data if type(previous_data) == dict else dict(),
            )
        else:
            ret = {}
            for k, v in declaration.items():
                prev = None
                if type(previous_data) == dict:
                    prev = previous_data.get(k)
                ret[k] = eval_nodes(v, prev)
            return ret
    if type(declaration) == list:
        raise Exception(
            "lists are not supported in bumpkit manifests due to the possibility of the shift problem"  # noqa: E501
        )
    return declaration


def eval_nodes_key(declaration=None, previous_data=None, key=[]):
    if len(key) == 0:
        return eval_nodes_recursively(declaration, previous_data)
    ret = previous_data if type(previous_data) is dict else dict()
    ret[key[0]] = eval_nodes_key(
        declaration=declaration.get(key[0])
        if type(declaration) is dict
        else None,
        previous_data=previous_data.get(key[0])
        if type(previous_data) is dict
        else None,
        key=key[1:],
    )
    return ret


def eval_nodes(declaration=None, previous_data=None, keys=[]):
    listed_nodes = list_nodes(declaration, previous_data)
    _keys = []
    for key in keys:
        if listed_nodes.get(key) is not None:
            _keys.append(key)
        else:
            for listed in listed_nodes:
                if listed.startswith(key):
                    _keys.append(listed)

    if len(_keys) == 0:
        keys = list(listed_nodes.keys())
    else:
        keys = _keys

    ordered_keys = keys  # older updates first
    ordered_keys.sort(key=lambda x: listed_nodes.get(x) or 0)

    ret = previous_data if type(previous_data) is dict else dict()
    for key in ordered_keys:
        keylist = key.split(".")
        try:
            ret = eval_nodes_key(declaration, ret, keylist)
        except KeyboardInterrupt:
            return ret
    return ret
